#!/usr/bin/env python3
"""init app

"""
from importlib import import_module
import inspect
import logging
from logging.handlers import RotatingFileHandler
import os
from typing import List
from uuid import uuid4


from flask import Flask, g, json
# MethodViewType is removed in flask 2.x
try:
    from flask.views import MethodViewType as MethodView
except ImportError:
    from flask.views import MethodView
from werkzeug.exceptions import HTTPException

from conf import app_config


basedir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
api_dir = os.path.abspath(os.path.dirname(__file__))


class BaseConfig:
    """project config
    
    """
    SECRET_KEY = app_config.sk


def create_app():
    """create flask app

    :return:
    """
    app = Flask('easy-flask')
    app.config.from_object(BaseConfig)

    register_logging(app)
    register_apis(app)

    @app.before_request
    def setup():
        """set different log_id for each request

        :return:
        """
        g.log_id = uuid4().hex
        g.logger = logging.LoggerAdapter(app.logger, extra={'log_id': g.log_id})

    @app.errorhandler(HTTPException)
    def handle_exception(e):
        """Return JSON instead of HTML for HTTP errors."""
        # start with the correct headers and status code from the error
        response = e.get_response()
        # replace the body with JSON
        response.data = json.dumps({
            "errno": e.code,
            "msg": "[%s] %s" % (e.name, e.description),
            "log_id": g.log_id
        })
        response.content_type = "application/json"
        return response

    return app


def register_logging(app: Flask):
    logging.logThreads = 0
    logging.logProcesses = 0
    logging.logMultiprocessing = 0
    formatter = logging.Formatter('%(asctime)s [%(module)s] %(levelname)s log_id[%(log_id)s] %(message)s')
    file_handler = RotatingFileHandler('logs/app.log', maxBytes=10 * 1024 * 1024, backupCount=10)
    file_handler.setFormatter(formatter)
    log_level = logging.DEBUG if app.debug else logging.INFO
    file_handler.setLevel(log_level)
    app.logger.setLevel(log_level)
    app.logger.addHandler(file_handler)
    
    
def add_resource(app: Flask, resource: MethodView, urls: List[str]):
    for url in urls:
        name = resource.__name__.lower() + url.replace('/', '_')
        app.add_url_rule(url, view_func=resource.as_view(name))


def register_apis(app: Flask):
    """register api from files from api dir

    :param api:
    :return:
    """
    api_file_data = os.walk(api_dir)
    for r, _, files in api_file_data:
        imp_path_list = r.replace(basedir, '')[1:].split(os.path.sep)
        imp_path = '.'.join(imp_path_list)
        for f in files:
            if f.endswith('api.py'):
                name = f.replace('.py', '')
                imp_file = '{}.{}'.format(imp_path, name)
                module = import_module(imp_file)
                class_list = inspect.getmembers(module, inspect.isclass)
                for c in class_list:
                    if type(c[1] == MethodView) and c[0] != 'Resource':
                        add_resource(app, c[1], c[1].uri)



#!/usr/bin/env python3
"""load app

"""
import os

from api import create_app


if not os.path.exists('logs'):
    os.mkdir('logs')


app = create_app()


if __name__ == '__main__':
    app.run(host='0.0.0.0', port=8000, debug=True, threaded=True)



#!/usr/bin/env python3
"""parser
Usage, for example:

request:
GET /api?name=xyz&age=10&extra={\"likes\": \"a,b,c,d\"}
parse to data ==>
{
    "name": "xyz",
    "age": 10,
    "extra": {
        "class": "a",
        "likes": ["a", "b", "c", "d"]
    }
}

# -*- code -*-
from flask import request
from easy_flask.utils.parser import parser, Required, Type, Min, Max, Split, Pattern, In, Default

extra_pattern = {
    "class": {Type: str, Default: "a", In: ["a", "b", "c"]},
    "likes": {Type: list, Split: ","}
}
pattern = {
    "name": {Type: str, Required: True},
    "age": {Type: int, Required: True, Min: 1, Max: 200},
    "extra": {Type: dict, Required: True}
}

data, err = parser(data=request.args.to_dict(), pattern=pattern)
...
"""
import json
from json.decoder import JSONDecodeError
from typing import Union


Type = 'type'  # param type: str, int, float, dict, list, bool
Required = 'required'  # if param required, set True
Default = 'default'  # set default value if param is None
Min = 'min'  # check min if param type is int or float
Max = 'max'  # check max if param type is int or float
Split = 'split'  # auto split str to list
In = 'in'  # check value in
Not = 'not'  # check value not
Pattern = 'pattern'  # if Type is dict or list, parse pattern for dict or each item(dict) in list


def parse_to_int(s, minimum: Union[None, int] = None, maximum: Union[None, int] = None):
    """parse unknown data to int, return None when exception

    :param s:
    :param minimum: check min if set
    :param maximum: check max if set
    :return:
    """
    try:
        tmp = int(s)
        if minimum is not None and tmp < minimum:
            return None
        if maximum is not None and tmp > maximum:
            return None
        return tmp
    except (ValueError, TypeError):
        return None


def parse_to_float(s, minimum: Union[None, float] = None, maximum: Union[None, float] = None):
    """parse unknown data to float, return None when exception

    :param s:
    :param minimum: check min if set
    :param maximum: check max if set
    :return:
    """
    try:
        tmp = float(s)
        if minimum is not None and tmp < minimum:
            return None
        if maximum is not None and tmp > maximum:
            return None
        return tmp
    except (ValueError, TypeError):
        return None


def parse_to_str(s):
    """parse object to str

    :param s:
    :return:
    """
    if isinstance(s, (dict, list)):
        return json.dumps(s, ensure_ascii=False)
    else:
        return str(s)


def json_loads(s):
    """load json from string, return None when exception

    :param s:
    :return:
    """
    if isinstance(s, (dict, list)):
        return s
    try:
        return json.loads(s)
    except (JSONDecodeError, TypeError):
        return None


def parser(data: dict, pattern: dict, remove_redundant: bool = False):
    """get and parse request data

    :param data:
    :param pattern: use to define attrs of each params,
        eg: {
                "name": {"type": str, "required": true},
                "nickname": {"type": str, "default": "zhangsan"},
                "age": {"type": int, "required": true, "min": 1, "max": 200}
            }
    :param remove_redundant: remove keys in data but not in pattern
    :return:
    """
    for k, v in pattern.items():
        t = v.get(Type)
        if t is None:
            continue
        # parse request data
        if data.get(k):
            if t == int:
                data[k] = parse_to_int(data[k], minimum=v.get(Min), maximum=v.get(Max))
                if data[k] is None:
                    return None, 'key[%s] parse int error, or check min|max error' % k
            elif t == float:
                data[k] = parse_to_float(data[k], minimum=v.get(Min), maximum=v.get(Max))
                if data[k] is None:
                    return None, 'key[%s] parse float error, or check min|max error' % k
            elif t == dict:
                data[k] = json_loads(data[k])
                if v.get(Pattern):
                    if not data[k]:
                        return None, 'key[%s] parse dict null or empty' % k
                    tmp, err = parser(data=data[k], pattern=v[Pattern])
                    if err:
                        return None, 'key[%s] parse dict error: %s' % (k, err)
            elif t == str:
                tmp = parse_to_str(data[k])
                if v.get(Split):
                    tmp = tmp.split(v[Split])
                data[k] = tmp
            elif t == list:
                if v.get(Split) and isinstance(data[k], str):
                    data[k] = data[k].split(v[Split])
                else:
                    data[k] = json_loads(data[k])
                    if v.get(Pattern):
                        if not isinstance(data[k], list):
                            return None, 'key[%s] parse list error, not list' % k
                        if len(data[k]) > 0:
                            tmp_list = []
                            for item in data[k]:
                                tmp, err = parser(data=item, pattern=v[Pattern])
                                if err:
                                    return None, 'key[%s] parse list error, item parse error: %s' % (k, err)
                                tmp_list.append(tmp)
                            data[k] = tmp_list
            elif t == bool:
                if data[k] in ['true', 'True', 'TRUE', 1, '1']:
                    data[k] = True
                else:
                    data[k] = False
            if data[k] is None:
                return None, 'parse[%s] error' % k

        # check required param
        if v.get(Required) and data.get(k) is None:
            return None, '%s is required' % k

        # check in
        if v.get(In):
            if t == list:
                if set(data[k]).difference(v[In]):
                    return None, 'key[%s] In check error' % k
            else:
                if data[k] not in data[In]:
                    return None, 'key[%s] In check error' % k
        # check not
        if v.get(Not):
            if t == list:
                if set(data[k]) & v[Not]:
                    return None, 'key[%s] Not check error' % k
            else:
                if data[k] in data[In]:
                    return None, 'key[%s] Not check error' % k

        # set default
        if data.get(k) is None and v.get(Default) is not None:
            data[k] = v[Default]

    # remove redundant in data but not in pattern
    if remove_redundant:
        redundant_keys = set(data.keys()).difference(set(pattern.keys()))
        for rk in redundant_keys:
            del data[rk]

    return data, None



#!/usr/bin/env python3
"""load conf

"""
import configparser
import os
import multiprocessing


conf_dir = os.path.abspath(os.path.dirname(__file__))
conf_file = os.path.join(conf_dir, 'config.ini')


class BaseConfig(object):

    def __init__(self, section: str):
        self._section = section

    def _load_conf(self):
        config = configparser.ConfigParser()
        config.read(conf_file, encoding='utf-8')

        for k, v in config.items(self._section):
            if hasattr(self, k):
                default_val = getattr(self, k)
                try:
                    if isinstance(default_val, int):
                        v = int(v)
                    elif isinstance(default_val, float):
                        v = float(v)
                    elif isinstance(default_val, bool):
                        v = True if str(v).lower() == 'true' else False
                except ValueError:
                    v = default_val
            setattr(self, k, v)
        return


class AppConfig(BaseConfig):

    def __init__(self):
        super().__init__(section='app')
        self.host = '0.0.0.0'
        self.port = 8000
        self.workers = multiprocessing.cpu_count() * 2 + 1
        self.threads = 2
        self.loglevel = 'info'
        self.worker_class = 'tornado'
        self.timeout = 60
        self._load_conf()


app_conf = AppConfig()



#!/usr/bin/env python3
"""define error info here

"""


class BaseError(object):
    errno = 1
    code = 500
    error_info = 'error'

    def __init__(self, msg: str = ''):
        msg = self.error_info + ':' + msg if msg else self.error_info
        self.msg = msg

    def __str__(self):
        return self.msg


class InternalServerError(BaseError):
    errno = 1
    code = 500
    error_info = 'internal server error'


class ParamsError(BaseError):
    errno = 2
    code = 240
    error_info = 'params error'



"""easy_flask

"""
import argparse
import shutil
import re
import os
import sys
from fnmatch import filter
from os.path import isdir, join


__version__ = '1.1.3'


def init_args():
    """init args

    :return:
    """
    description = """easy-flask, use it to make a project quickly. 
    """
    parser = argparse.ArgumentParser(description=description)
    parser.add_argument('-n', '--new', type=str, help='input project name, generate a flask project')

    args = parser.parse_args()
    return args


def check_project_name(name: str):
    """project name example: flask_demo

    :param name:
    :return:
    """
    match = re.match('^[a-zA-z]\w+', name)
    if not match:
        return False
    if name != match.group():
        return False
    return True


def get_path():
    path = ''
    for p in sys.path:
        if not p.endswith('-packages'):
            continue
        path = os.path.join(p, 'easy_flask')
        if os.path.exists(path):
            break
    return path


def include_patterns(*patterns):
    """Factory function that can be used with copytree() ignore parameter.

    Arguments define a sequence of glob-style patterns
    that are used to specify what files to NOT ignore.
    Creates and returns a function that determines this for each directory
    in the file hierarchy rooted at the source directory when used with
    shutil.copytree().
    """
    def _ignore_patterns(path, names):
        keep = set(name for pattern in patterns for name in filter(names, pattern))
        ignore = set(name for name in names if name not in keep and not isdir(join(path, name)))
        return ignore
    return _ignore_patterns


def main():
    args = init_args()
    p_name = args.new
    if check_project_name(args.new) is False:
        print('project name invalid(%s), eg: easy_flask' % p_name)
        sys.exit(1)

    curr_dir = os.getcwd()
    src_dir = get_path()
    if not src_dir:
        print('ERROR: cannot find easy-flask in python packages')
        sys.exit(1)
    
    dest_dir = os.path.join(curr_dir, p_name)
    shutil.copytree(src_dir, dest_dir, ignore=include_patterns('*.py', '*.ini', '*.sh'))
    
    useless_file = os.path.join(dest_dir, '__init__.py')
    os.remove(useless_file)
    
    print('flask project create success, enjoy~')
    sys.exit(0)



#!/usr/bin/env python3
"""demo api

"""
from flask import g

from internal import Resource
from utils.parser import Type, Required, Default, Min, Max


class DemoApi(Resource):

    uri = ['/api/demo']

    def get(self):
        pattern = {
            'num': {Type: int, Min: 10, Max: 100},
            'print': {Type: str, Required: True},
            'default': {Type: str, Default: 'demo'}
        }

        data, err = self.parse_request_data(pattern=pattern)
        if err:
            g.logger.warning(err)
            return err
        g.logger.info('success')
        return {'result': data}
